/****************************************************************
*			lz77 packer v1.2			*
*****************************************************************
* Source adapted and optimized to/for Atari 16/32bit machines	*
* by ray//.tSCc. 2003. Final bugfixes 01.04.05		*
*								*
* Base framework: Haruhiko Okumura				*
*****************************************************************
* eml:	 ray@tscc.de						*
* http://ray.tscc.de						*
*****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifdef __GNUC__
# include <netinet/in.h>
#else	/* assume big-endian i.e. host->network byteorder conversion is no-op */
# define ntohl(x) x
# define htonl(x) x
#endif

#ifdef USE_ASM
extern void d_lz77(unsigned char *src, unsigned char *dst);
#endif

#define SHOW_PROGRESS 0		/* whether to show encode/decode progress */
#define DEBUG 0			/* whether to check for bounds & output extra info */
#if DEBUG
# define DPRINT(args) printf args
#else
# define DPRINT(args)
#endif

#define N		4096	/* Size of dictonary */
#define F		17	/* Upper limit for match length */
#define THRESHOLD	2	/* Encode strings longer than this */
							 
#define NIL		N	/* Index for root of binary search trees */

static unsigned long int
	textsize = 0,		/* Text size counter */
	codesize = 0;		/* Code size counter */

static unsigned char text_buf[N+F-1];	/* Ring buffer for string comparison */ 

static int match_position, match_length;	/* Set by the InsertNode() procedure. */
static int lson[N+1],rson[N+257], dad[N+1];	/* Left & right children & parents (binary search trees) */



/*******************************************
*
*  void InitTree(void)
*
* Initialise binary search trees for fast
* string comparison
*
*******************************************/

static void InitTree(void)
{
	int  i;

	for (i=N+1;i<=N+256;i++) rson[i] = NIL;
	for (i=0;i<N;i++) dad[i] = NIL;
}


/*******************************************
*
*  void InsertNode(int r)
*
* Insert text_buf[r..r+F-1], into one of the trees and return
* the longest-match position and length.
*
* If match_length = F, then removes the old node.
*
*******************************************/

static void InsertNode(int r)
{
	int  i, p, cmp;
	unsigned char	*key;

	cmp = 1;
	key = &text_buf[r];
	p = N+1+key[0];

	rson[r] = lson[r] = NIL;
	match_length = 0;
	for (;;) {
		if (cmp >= 0) {
			if (rson[p] != NIL)
				p = rson[p];
			else {
				rson[p] = r;
				dad[r] = p;
				return;
			}
		}
		else {
			if (lson[p] != NIL)
				p = lson[p];
			else {
				lson[p] = r;
				dad[r] = p;
				return;
			}
		}
		for (i = 1; i < F; i++)
			if ((cmp = key[i] - text_buf[p + i]) != 0)
				break;
		if (i > match_length) {
			match_position = r-p;
			if ((match_length = i) >= F)
				break;
		}
	}
	dad[r] = dad[p];
	lson[r] = lson[p];
	rson[r] = rson[p];

	dad[lson[p]] = r;
	dad[rson[p]] = r;
	if (rson[dad[p]] == p)
		rson[dad[p]] = r;
	else
		lson[dad[p]] = r;
	dad[p] = NIL;	/* remove p */
}


/*******************************************
*
*  void DeleteNode(int p)
*
* InsertNode's counterpart
*
*******************************************/

static void DeleteNode(int p)
{
	int  q;

	if (dad[p] == NIL)
		return;  /* not in tree */
	if (rson[p] == NIL)
		q = lson[p];
	else if (lson[p] == NIL)
		q = rson[p];
	else {
		q = lson[p];
		if (rson[q] != NIL) {
			do {
				q = rson[q];
			} while (rson[q] != NIL);
			rson[dad[q]] = lson[q];
			dad[lson[q]] = dad[q];
			lson[q] = lson[p];
			dad[lson[p]] = q;
		}
		rson[q] = rson[p];
		dad[rson[p]] = q;
	}
	dad[q] = dad[p];
	if (rson[dad[p]] == p)
		rson[dad[p]] = q;
	else
		lson[dad[p]] = q;

	dad[p] = NIL;
}



/*******************************************
*
*  void Progress(void)
*
* Report progress of the en/decoding procedure
*
*******************************************/

#if SHOW_PROGRESS
static void Progress(void)
{
	static a = 0;
	const char progress[4] = {'|','/','-','\\'};

	printf("%c\b",progress[a++]);
	a &= 3;
	
	return;
}
#else
#define Progress()
#endif

static unsigned long get_size(FILE *infile)
{
	unsigned long fl=0;
	fseek(infile, 0, SEEK_END);
	fl = ftell(infile);	/* inputfile length */
	rewind(infile);
	return fl;
}

/*******************************************
*
*  void Encode(void)
*
* Encoding procedure. Return zero for success.
*
*******************************************/

static int Encode(FILE *infile, FILE *outfile)
{
	int  i,c,len,r,s,last_match_length,code_buf_ptr;
	unsigned char	code_buf[17], mask;
	unsigned long fl;

	fl = htonl(get_size(infile));
	fwrite(&fl,sizeof(fl),1,outfile);

	InitTree();		/* Initialize trees */
	code_buf[0] = 0xff;	/* Initialise header flags for the first time */

	code_buf_ptr = 1;
	mask = 0x80;	/* Headerbit to be masked out in case of a literal string */
	
	r = N - F;
	for (i = 0; i < r && (c = getc(infile)) != EOF; i++) text_buf[i] = c^0xff; /* Clear the buffer */

	/* Read F bytes into the last F bytes of the buffer */
	rewind(infile);
	for (len = 0; len < F && (c = getc(infile)) != EOF; len++)
		text_buf[r + len] = c;
			
	if ((textsize = len) == 0)
		return 0;

	for (i = 1; i <= F; i++)
		InsertNode(r - i);	/* Insert the F strings into the tree */
	InsertNode(r);		/* Finally, insert the whole string. */
	
	/* Encoding loop */	
	s = 0;
	do {
		if (match_length > len)
			match_length = len;
		
		/* match is too short or exceeds file entry */
		if ((match_length <= THRESHOLD) || ((ftell(outfile)&4095) < match_position)) {
			match_length = 1;
			code_buf[0] -= mask;  			/* Mask out headerbit */
			code_buf[code_buf_ptr++] = text_buf[r]; /* Send literally */
			
		/* otherwise we found a match, so encode it */
		} else {

			code_buf[code_buf_ptr++] = (unsigned char)
				(((match_position >> 4) & 0xf0)
				| (match_length - THRESHOLD));  /* Send index/length pair */
   			code_buf[code_buf_ptr++] = (unsigned char) match_position;
		}
		
		if ((mask >>= 1) == 0) {  /* Shift mask right one bit. */
		
			for (i = 0; i < code_buf_ptr; i++)	/* Send at most 8 units */
				putc(code_buf[i], outfile);
				
			codesize += code_buf_ptr;		/* Reset encoding variables */
			code_buf[0] = 0xff;
			code_buf_ptr = 1;
			mask = 0x80;
		}
		
		last_match_length = match_length;
		
		for (i = 0; i < last_match_length && ((c = getc(infile)) != EOF); i++) {
		
			DeleteNode(s);			/* Delete old strings and */
			text_buf[s] = c;		/* read on */

			/* Speedup for strings near the end of the input */
			if (s < F - 1)
				text_buf[s + N] = c;
								
			s = (s+1)&(N-1);	/* Keep the ringbuffer's range */
			r = (r+1)&(N-1);

			InsertNode(r);		/* Register the string in text_buf[r..r+F-1] */
		}

		textsize += i;
		Progress();
		while (i++ < last_match_length) {	/* Handle segments after the end of */
			DeleteNode(s);			/* the input */
			s = (s+1)&(N-1);
			r = (r+1)&(N-1);
			if (--len)
				InsertNode(r);
		}
		
	} while (len > 0);
	
	if (code_buf_ptr > 1) {		/* Send remaining code. */
   		for (i = 0; i < code_buf_ptr; i++)
			putc(code_buf[i], outfile);
		codesize += code_buf_ptr;
	}

	if (mask == 0x80) {		/* Write the EOF flag *bugfix 2004*/
		c = 0x80; putc(c,outfile);
	}
	c = 0x00; putc(c,outfile);

	printf("Achieved ratio: %lu%%\n", (codesize*100+textsize/2) / textsize);
	return 0;
}



/*******************************************
*
*  void Decode(void)
*
* Encode's counterpart. Return zero for success.
*
*******************************************/

static int Decode(FILE *infile, FILE *outfile)
{
#ifndef USE_ASM
	unsigned int tag;
	int  i, j, r, c, loop1;
	int idxin = 0, idxout = 0;
#endif
	/* alloc buffers */
	unsigned char *inbuf, *outbuf;
	long sizein, sizeout, fl;

	fread(&fl, sizeof(fl), 1, infile);
	sizeout = ntohl(fl);
	outbuf = malloc(sizeout);
	sizein = get_size(infile);
	inbuf = malloc(sizein);
	fseek(infile, sizeof(fl), SEEK_SET);
	fread(inbuf, sizein, 1, infile);

#if USE_ASM
	DPRINT(("calling d_lz77(0x%p, 0x%p);\n", inbuf, outbuf));
	d_lz77(inbuf, outbuf);
	fwrite(outbuf, sizeout, 1, outfile);
	return 0;
#else

 	r = N - F;
 	
	for (;;) {
		tag = inbuf[idxin++];		/* Load command header */

		for (loop1 = 0; loop1 < 8; loop1++) {
	
			/* Decode packed string segment */
			if (tag & 0x80) {
				
				/* End of input stream? */
#if DEBUG
				if (idxin >= sizein) {
						printf("ABORT: output overrun.\n");
						DPRINT(("idxin: %d, sizein: %ld\n", idxin, sizein));
						return 1;
				}
#endif
				if ((j = inbuf[idxin++]) == 0) {
					fwrite(outbuf, sizeout, 1, outfile);
					DPRINT(("idxin: %d, sizein: %ld\n", idxin, sizein));
					return 0;
				}

				i = inbuf[idxin++];

				i |= (j<<4)&0xf00;	/* Compute position & length of string */
				j &= 0x0f;

				i = (r-i)&(N-1);

				while(j != -2) {
					c = text_buf[(i++)&(N-1)]; /* Copy from dictonary */
#if DEBUG
					if (idxout >= sizeout) {
						printf("ABORT: output overrun.\n");
						DPRINT(("idxin: %d, sizein: %ld\n", idxin, sizein));
						return 1;
					}
#endif
					outbuf[idxout++] = c;
					text_buf[r++] = c;
					r &= (N-1);
					j--;
					Progress();
				}

				/* Restore literal units */	
			} else {
#if DEBUG
				if (idxin >= sizein || idxout >= sizeout) {
					printf("ABORT: input/output overrun.\n");
					DPRINT(("idxin: %d, sizein: %ld\n", idxin, sizein));
					return 1;
				}
#endif
				c = inbuf[idxin++];
				outbuf[idxout++] = c;
				text_buf[r++] = c;
				r &= (N-1);
				Progress();
			}
			tag <<= 1;	/* Next entry */
		}
	}
	return 1;
#endif
}

/* TOS specific way for timing compression/uncompression time */
#if defined(__ATARI_ST__) || defined(__MINT__) || defined(__VBCC__)
# ifdef __VBCC__
#  include <tos.h>
# else
#  include <osbind.h>
# endif
static long clock_value;
static long clock_set(void) { return clock_value = *(long *)1210L; }
static long clock_get(void) { Supexec(clock_set); return clock_value; }

static long timer_value;
static void timer_start(void) { timer_value = clock_get(); }
static void timer_print(void) { printf("...in %d ms\n", (int)((clock_get() - timer_value) * 5)); }
#else
# define timer_start()
# define timer_print()
#endif

int main(int argc, char *argv[])
{
	FILE *infile, *outfile;
	int ret;
	char *s;

	printf("LZ77 packer  by ray//.tSCc. 2003/2004\n\n");
	
	/* Check cdl */
	if ((argc != 4) || (s=argv[1], s[1] || strpbrk(s, "pPxX") == NULL) ) {
		printf("Usage: lz77 switch infile.ext outfile.ext\n\n");
		printf("Valid switches are:\n");
		printf("  p (compress input file)\n");
		printf("  x (extract input file)\n\n");
		return 0;
	}
	
	if ((s = argv[2], (infile  = fopen(s, "rb")) == NULL)||
	    (s = argv[3], (outfile = fopen(s, "wb")) == NULL)) {
		printf("Failed to open %s\n", s);  return 0;
	}
	timer_start();
	if (toupper(*argv[1]) == 'P') {
		printf("Deflating %s: ", argv[2]);
		ret = Encode(infile, outfile);
	} else {
		printf("Inflating %s: ", argv[2]);
		ret = Decode(infile, outfile);
	}
	timer_print();

	if (ret == 0) {
		printf("Ok.\n");
	} else {
		printf("Fail.\n");
	}
	fclose(outfile);
	fclose(infile);
	return ret;
}
